package liv.jmxmonitor;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.DynamicMBean;
import javax.management.InvalidAttributeValueException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanOperationInfo;
import javax.management.MBeanParameterInfo;

/**
 * A class which maps integer to their frequency and keeps a running total of
 * all numbers added to the map. Every time a number is added to the map, if
 * already present, its frequency is increased and the value added to the
 * running total. Please note that this class <b>DOES NOT</b> implement the
 * <code>Map</code> interface. This class is synchronized.
 *
 * @author Liviu Tudor http://about.me/liviutudor
 */
public final class FreqMap implements DynamicMBean {
    /* These are all names of attributes and operations to be used with JMX */
    private static final String   ATTR_TOTAL       = "total";
    private static final String   ATTR_SIZE        = "size";
    private static final String   OP_CLEAR         = "clear";
    private static final String   OP_GET_FREQUENCY = "getFrequency";
    private static final String   OP_ADD           = "add";

    /**
     * Stores the frequency map. The underlying implementation is a
     * <code>HashMap</code>.
     */
    private Map<Integer, Integer> frequency;

    /**
     * Stores the total of all numbers added to the map.
     */
    private int                   total;

    /**
     * Used internally to synchronize read/write access.
     */
    private ReadWriteLock         lock;

    /**
     * Default constructor. Initializes an empty frequency map and prepares the
     * locks.
     */
    public FreqMap() {
        frequency = new HashMap<Integer, Integer>();
        total = 0;
        lock = new ReentrantReadWriteLock();
    }

    /**
     * Adds a number to our frequency map. This also updates the running total.
     *
     * @param number
     *            number to be added to our freq map and running total
     */
    public void add(int number) {
        lock.writeLock().lock();
        try {
            internalUnsyncAdd(number);
        } finally {
            lock.writeLock().unlock();
        }
    }

    /**
     * Provides an <b>atomic</b> way of adding multiple numbers to the frequency
     * map in one go.
     *
     * @param number
     *            Array of numbers to add
     */
    public void add(int number[]) {
        if (number == null || number.length == 0) // nothing to do!
            return;
        lock.writeLock().lock();
        try {
            for (int i = 0; i < number.length; i++)
                internalUnsyncAdd(number[i]);
        } finally {
            lock.writeLock().unlock();
        }
    }

    /**
     * Used internally to perform the necessary steps to add an item to the map
     * and total. This is done in an un-synchronized manner so the caller has to
     * employ synchronization mechanisms to the map and total before calling
     * this.
     *
     * @param number
     *            number to be added to our freq map and running total
     */
    private void internalUnsyncAdd(int number) {
        int count = 0;
        if (frequency.containsKey(number)) count = frequency.get(number);
        count++;
        frequency.put(number, count);
        total += number;
    }

    /**
     * Retrieves the running total.
     *
     * @return Current value of {@link total}
     */
    public int getTotal() {
        lock.readLock().lock();
        try {
            return total;
        } finally {
            lock.readLock().unlock();
        }
    }

    /**
     * Retrieves the frequency of the given number.
     *
     * @param number
     *            Integer to retrieve the frequency for.
     * @return number of times the number was added to the map or zero if the
     *         number is not found in our frequency map (in other words it was
     *         never added).
     */
    public int getFrequency(int number) {
        lock.readLock().lock();
        try {
            if (!frequency.containsKey(number)) return 0;
            return frequency.get(number);
        } finally {
            lock.readLock().unlock();
        }
    }

    /**
     * Resets the frequency map and the running total. Basically empties the map
     * and sets the running total to zero.
     */
    public void clear() {
        lock.writeLock().lock();
        try {
            frequency.clear();
            total = 0;
        } finally {
            lock.writeLock().unlock();
        }
    }

    /**
     * Retrieves the size of our frequency map.
     *
     * @return the number of <b>distinct</b> numbers added to the map.
     */
    public int size() {
        lock.readLock().lock();
        try {
            return frequency.size();
        } finally {
            lock.readLock().unlock();
        }
    }

    @Override
    public Object getAttribute(String attr) throws AttributeNotFoundException {
        if (ATTR_SIZE.equals(attr))
            return size();
        else if (ATTR_TOTAL.equals(attr)) return getTotal();
        throw new AttributeNotFoundException(attr);
    }

    @Override
    public AttributeList getAttributes(String[] attributes) {
        if (attributes == null) return null;
        if (attributes.length == 0) return new AttributeList();
        AttributeList list = new AttributeList();
        for (String s : attributes) {
            try {
                list.add(new Attribute(s, getAttribute(s)));
            } catch (AttributeNotFoundException e) {
                e.printStackTrace();
            }
        }
        return list;
    }

    @Override
    public MBeanInfo getMBeanInfo() {
        MBeanAttributeInfo[] attributes = new MBeanAttributeInfo[] {
                new MBeanAttributeInfo(ATTR_SIZE, int.class.getCanonicalName(),
                        "Number of distinct numbers added to the map so far", true, false, false),
                new MBeanAttributeInfo(ATTR_TOTAL, int.class.getCanonicalName(),
                        "Running total of all the numbers ever added to the map", true, false, false) };

        MBeanOperationInfo[] operations = new MBeanOperationInfo[] {
                new MBeanOperationInfo(OP_CLEAR, "Resets the frequency map", null, void.class.getCanonicalName(),
                        MBeanOperationInfo.ACTION),
                new MBeanOperationInfo(OP_ADD, "Adds a number to the map",
                        new MBeanParameterInfo[] { new MBeanParameterInfo("number", int.class.getCanonicalName(),
                                "Number to be added") }, void.class.getCanonicalName(), MBeanOperationInfo.ACTION),
                new MBeanOperationInfo(OP_GET_FREQUENCY, "Retrieves the frequency of a number",
                        new MBeanParameterInfo[] { new MBeanParameterInfo("number", int.class.getCanonicalName(),
                                "Number to retrieve frequency for") }, int.class.getCanonicalName(),
                        MBeanOperationInfo.INFO) };
        MBeanInfo info = new MBeanInfo(this.getClass().getCanonicalName(), "Frequency map", attributes, null,
                operations, null);
        return info;
    }

    @Override
    public Object invoke(String actionName, Object[] params, String[] signature) throws MBeanException {
        if (OP_CLEAR.equals(actionName)) {
            clear();
            return null;
        } else if (OP_ADD.equals(actionName)) {
            int param = (Integer) params[0];
            add(param);
            return null;
        } else if (OP_GET_FREQUENCY.equals(actionName)) {
            int param = (Integer) params[0];
            return getFrequency(param);
        } else
            throw new MBeanException(new IllegalArgumentException("No such method " + actionName));
    }

    @Override
    public void setAttribute(Attribute attr) throws InvalidAttributeValueException {
        /* we don't support setting any attributes */
        throw new InvalidAttributeValueException("Cannot set attribute " + attr.getName());
    }

    @Override
    public AttributeList setAttributes(AttributeList attributes) {
        /* we don't support setting any attributes */
        return null;
    }
}
